{$A+,B-,D+,E-,F-,G+,I+,L+,N+,O-,P-,Q-,R-,S-,T-,V+,X+,Y+}
{by Behdad Esfahbod}
unit
  _Family;

interface

  procedure ReadRelations;
  procedure ReadFamily;
  procedure ReadQueries;

  procedure GenOutput;

implementation

uses
  _Global, _Person, _Relation;

var
  People : TPeopleArray;
  Relations : TRelationsArray;
  Queries : array [1 .. MaxPersons] of Integer;
  Id : array [1 .. MaxPersons] of Integer;
  Path, PathRel, G : array [1 .. MaxPersons, 1 .. MaxPersons] of Byte;
  N, M, Q : Integer;

procedure IranSystemFont; external; {$l iransys.obj}

procedure SetFarsiFont;
begin
  asm
    mov ax, 1110h
    mov cx, 256
    mov dx, 0
    mov bl, 0
    mov bh, 16
    push bp
    mov bp, offset IranSystemFont
    push seg IranSystemFont
    pop es
    int 10h
    pop bp
  end;
end;

procedure CheckFarsi (S : string);
var
  I : Integer;
begin
  if S = '' then Exit;
  if not Farsi then
  begin
    for I := Length(S) downto 0 do
      if S[I] >= #128 then
        Break;
    if I = 0 then Exit;
    Farsi := True;
    SetFarsiFont;
  end;
end;

function  FindQueryPerson (S : TName) : Integer;
var
  I, J : Integer;
  T : TName;
begin
  Val(S, J, I);
  if I = 0 then
  begin
    for I := 1 to N do
      if Id[I] = J then
      begin
        FindQueryPerson := I;
        Exit;
      end;
    Error('person ' + S + ' not introduced yet.');
  end
  else
  begin
    T := UCase(S);
    for I := 1 to N do
      if (People[I] <> nil) and (UCase(People[I]^.Name) = T) then
      begin
        FindQueryPerson := I;
        Exit;
      end;
    Error(S + ' not introduced yet.');
  end;
  FindQueryPerson := 0;
end;

function  FindPerson (S : TName) : PPerson;
begin
  FindPerson := People[FindQueryPerson(S)];
end;

function  FindRelation (S : TName) : PRelation;
var
  I, J : Integer;
  T : TName;
begin
  T := UCase(S);
  for I := 1 to M do
    if (Relations[I] <> nil) and (UCase(Relations[I]^.Name) = T) then
    begin
      FindRelation := Relations[I];
      Exit;
    end;
  Error('relation ' + S + ' not defined yet.');
  FindRelation := nil;
end;

procedure ReadFamily;
var
  F : Text;
  S : string;
  S1, S2, SName, SSureName : TName;
  P1, P2 : PPerson;
  X : TSex;
  I, J : Integer;
begin
  Assign(F, FamilyFile);
  Reset(F);
  while not Eof(F) do
  begin
    Readln(F, S);
    if S = '' then
      Continue;

    S := S + ' ';
    if S[1] in ['0' .. '9'] then
    begin
      {Add person}
      Inc(N);
      for I := 1 to Length(S) do if not (S[I] in ['0' .. '9']) then Break;
      Val(Copy(S, 1, I - 1), Id[N], J);
      if Id[N] = 0 then
      begin
        Dec(N);
        Break;
      end;
      for I := I to Length(S) do if S[I] in ['m', 'M', 'f', 'F'] then Break;
      if S[I] in ['m', 'M'] then X := Male else X := Female;
      for J := I + 1 to Length(S) do if S[J] <> ' ' then Break;
      for I := J to Length(S) do if S[I] = ' ' then Break;
      S1 := Copy(S, J, I - J);
      for J := I to Length(S) do if S[J] <> ' ' then Break;
      for I := J to Length(S) do if S[I] = ' ' then Break;
      SSureName := Copy(S, J, I - J);
      for J := I to Length(S) do if S[J] <> ' ' then Break;
      for I := J to Length(S) do if S[I] = ' ' then Break;
      SName := Copy(S, J, I - J);
      if SName = '' then
      begin
        SName := SSureName;
        SSureName := '';
      end;
      People[N] := New(PPerson, Init(S1, SName, SSureName, N, X));
    end
    else
    begin
      S := S + ' ';
      for I := 2 to Length(S) do if S[I] <> ' ' then Break;
      for J := I to Length(S) do if S[J] = ' ' then Break;
      S1 := Copy(S, I, J - I);
      for I := J to Length(S) do if S[I] <> ' ' then Break;
      for J := I to Length(S) do if S[J] = ' ' then Break;
      S2 := Copy(S, I, J - I);
      P1 := FindPerson(S1);
      P2 := FindPerson(S2);
      case UpCase(S[1]) of
        'C': P2^.HasChild(P1);
        'P': P2^.HasParent(P1);
        'M': P2^.HasSpouse(P1);
      else
        Error('Invalid basic relation "' + S[1] + '".');
      end;
    end;
  end;
  Close(F);
end;

procedure ReadRelations;
var
  F : Text;
  S : string;
  S1, SFarsi : TName;
  S2, T : TDef;
  I, J : Integer;
begin
  Assign(F, RelationFile);
  Reset(F);
  while not Eof(F) do
  begin
    Readln(F, S);
    if S = '' then
      Continue;
    while S[Ord(S[0])] = ' ' do Dec(S[0]);
    I := Length(S);
    while S[I] >= #128 do
      Dec(I);
    SFarsi := Copy(S, I + 1, 255);
    CheckFarsi(SFarsi);
    Byte(S[0]) := I;
    while S[Ord(S[0])] = ' ' do Dec(S[0]);
    S := S + ' ';
    for I := 1 to Length(S) do if S[I] = ' ' then Break;
    S1 := Copy(S, 1, I - 1);
    for I := I to Length(S) do if S[I] <> ' ' then Break;
    S2 := Copy(S, I, 255);
    while S2[Length(S2)] = ' ' do
      Dec(S2[0]);
    if S2 = '' then
    begin
      SFarsi := '';
      S2 := '[' + S1 + ']';
      J := 2;
      I := Pos('-', S2);
      while  I <> 0 do
      begin
        if Farsi then
          SFarsi := FindRelation(Copy(S2, J, I - J))^.FarsiName + SFarsi;
        S2[I] := '[';
        Insert(']', S2, I);
        J := I + 1;
        I := Pos('-', S2);
      end;
      if Farsi then
        SFarsi := FindRelation(Copy(S2, J + 1, Length(S2) - J - 1))^.FarsiName+ SFarsi;
    end;
    CheckFarsi(SFarsi);
    I := Pos('[', S2);
    J := Pos(']', S2);
    while I + J <> 0 do
    begin
      if ((I + J <> 0) and (I * J = 0)) or (J < I) then
        Error(S1 + ': Unmatched "[" or "]".');
      T := S2;
      Delete(S2, I, J - I + 1);
      Insert(FindRelation(Copy(T, I + 1, J - I - 1))^.Definition, S2, I);
      I := Pos('[', S2);
      J := Pos(']', S2);
    end;
    Inc(M);
    Relations[M] := New(PRelation, Init(S1, SFarsi, M, S2));
  end;
  Close(F);
end;

procedure ReadQueries;
var
  F : Text;
  S : string;
  Mark : array [1 .. MaxPersons] of Boolean;
  I, J : Integer;
begin
  Assign(F, QueryFile);
  Reset(F);
  FillChar(Mark, SizeOf(Mark), 0);
  while not Eof(F) do
  begin
    Readln(F, S);
    Inc(Q);
    Queries[Q] := FindQueryPerson(S);
    Mark[Queries[Q]] := True;
  end;
  J := Q;
  for I := 1 to N do
    if not Mark[I] then
    begin
      Inc(J);
      Queries[J] := I;
    end;
  Close(F);
end;

function  WritePath (A, B : Integer) : string;
begin
  if Path[A, B] = 0 then
    WritePath := Relations[PathRel[A, B]]^.FarsiName
  else
    if Farsi then
      WritePath := WritePath(Path[A, B], B) + ' ' + WritePath(A, Path[A, B])
    else
      WritePath := WritePath(A, Path[A, B]) + ' of ' + WritePath(Path[A, B], B);
end;

procedure WritelnR (S : string);
begin
  Writeln(S : 78);
end;

procedure WriteRelation (A, B : Integer);
var
  S : string;
begin
  if (A <> B) and (G[A, B] <= N) then
    if Farsi then
      WritelnR('. ' + People[B]^.WriteName + ' '
               + WritePath(A, B) +
               ' ' + People[A]^.WriteName + ' --')
    else
      Writeln('-- ' + People[A]^.WriteName + ' is '
              + WritePath(A, B) +
              ' of ' + People[B]^.WriteName + '.');
end;

procedure GenOutput;
var
  P : PPerson;
  I, J, K, Num : Integer;
  S : string;

  procedure WriteQueryName;
  var
    K : Integer;
  begin
    if Farsi then
    begin
      Str(I, S);
      for K := 1 to Length(S) do
        Byte(S[K]) := Ord(S[K]) + Ord('') - Ord('0');
      WritelnR(':' + P^.WriteName + ' ' + S + '  ');
    end
    else
      Writeln('Query Number ', I, ', ' + P^.WriteName + ':');
  end;

begin
  Assign(Output, OutputFile);
  Rewrite(Output);
  FillChar(Path, SizeOf(Path), 0);
  FillChar(PathRel, SizeOf(PathRel), 0);
  for I := 1 to N do
    for J := 1 to N do
      G[I, J] := N + 1;
  for I := 1 to N do
    G[I, I] := 0;
  if Farsi then
    WritelnR(': ')
  else
    Writeln('Solving Queries:');
  Writeln;
  if Farsi then
    WritelnR(':    ')
  else
    Writeln('Listing People in Each Close Relationship:');
  for I := 1 to N do
  begin
    P := People[Queries[I]];
    if I <= Q then
    begin
      Writeln;
      WriteQueryName;
    end;
    for J := 1 to M do
      with Relations[J]^ do
      begin
        SolveFor(P, People, N);
        if Formal then
        begin
          Num := 0;
          for K := 1 to N do
            if InList(People[K]) then
            begin
              G[K, P^.Id] := 1;
              PathRel[K, P^.Id] := J;
              if People[K]^.Formal then
              begin
                if Farsi then
                  if Num > 0 then S := ' ' + S else S := ' = ' + FarsiName + ' --'
                else
                  if Num > 0 then S := S + ', ' else S := '-- ' + Name + ' = ';
                Inc(Num);
              end;
              if Farsi then
                S := People[K]^.WriteName + S
              else
                S := S + People[K]^.WriteName;
            end;
          if (Num > 0) and (I <= Q) then
            if Farsi then
              WritelnR(S)
            else
              Writeln(S);
        end;
      end;
  end;
  Writeln;
  if Farsi then
    WritelnR(':    ')
  else
    Writeln('Listing Closest Relation for Each Person:');
  Writeln;
  {Floyd-Warshall all-pairs-shortest-path algorithm}
  for K := 1 to N do
    for I := 1 to N do
      for J := 1 to N do
        if Integer(G[I, K]) + Integer(G[K, J]) < G[I, J] then
        begin
          G[I, J] := G[I, K] + G[K, J];
          Path[I, J] := K;
        end;
  for I := 1 to Q do
  begin
    P := People[Queries[I]];
    WriteQueryName;
    for J := 1 to N do
      if People[J]^.Formal then
      begin
        WriteRelation(J, P^.Id);
        WriteRelation(P^.Id, J);
      end;
    Writeln;
  end;
  Close(Output);
end;

begin
  Farsi := False;
end.
