{$A+,B-,D+,E-,F-,G+,I+,L+,N+,O-,P-,Q-,R-,S-,T-,V+,X+,Y+}
{by Behdad Esfahbod}
unit
  _Person;

interface

  uses
    _Global;

  type
    PPerson = ^ TPerson;
    PPeopleArray = ^ TPeopleArray;
    TPeopleArray = array [1 .. MaxPersons] of PPerson;
    TPerson = object

    private
      mName : TName;
      mFName : TName;
      mFSureName : TName;
      mId : TId;
      mSex : TSex;
      mFormal : Boolean;
      Parent, Spouse, Child, Sibling : PPerson;

    public
      constructor Init (TheName, TheFarsiName, TheFarsiSureName : TName;
                        TheId : TId; TheSex : TSex);

      procedure HasParent (P : PPerson);
      procedure HasSpouse (P : PPerson);
      procedure HasChild (P : PPerson);
      procedure List (var L : TList; Action : Char);

      function  Name : TName;
      function  FarsiName : TName;
      function  Id : TId;
      function  Sex : TSex;
      function  Formal : Boolean;
      function  WriteName : string;

      function  Son : PPerson;
      function  Daughter : PPerson;
      function  Father : PPerson;
      function  Mother : PPerson;
      function  Husband : PPerson;
      function  Wife : PPerson;

      function  IsMale : Boolean;
      function  IsFemale : Boolean;
      function  IsParent (P : PPerson) : Boolean;
      function  IsSpouse (P : PPerson) : Boolean;
      function  IsChild (P : PPerson) : Boolean;
      function  IsSon (P : PPerson) : Boolean;
      function  IsDaughter (P : PPerson) : Boolean;
      function  IsFather (P : PPerson) : Boolean;
      function  IsMother (P : PPerson) : Boolean;
      function  IsHusband (P : PPerson) : Boolean;
      function  IsWife (P : PPerson) : Boolean;
    end;

implementation

constructor TPerson.Init (TheName, TheFarsiName, TheFarsiSureName : TName;
                          TheId : TId; TheSex : TSex);
begin
  mFormal := TheName[1] <> '-';
  if TheName[1] = '-' then
    Delete(TheName, 1, 1);
  mName := TheName;
  mFName := TheFarsiName;
  mFSureName := TheFarsiSureName;
  mId   := TheId;
  mSex  := TheSex;
  Parent  := nil;
  Spouse  := nil;
  Child   := nil;
  Sibling := nil;
end;


procedure TPerson.HasParent (P : PPerson);
begin
  P^.HasChild(@Self);
end;

procedure TPerson.HasSpouse (P : PPerson);
begin
  if Sex = P^.Sex then
    Error(mName + ' and ' + P^.mName + ': homosexuality not supported.');
  if (Spouse <> nil) and (Spouse <> P) then
    Error(mName + ' has already married, multiple marriages not supported.');
  if (P^.Spouse <> nil) and (P^.Spouse <> @Self) then
    P^.HasSpouse(@Self);
  Spouse := P;
  P^.Spouse := @Self;
end;

procedure TPerson.HasChild (P : PPerson);
begin
  if Spouse = nil then
    Error(mName + ' has not married yet, can''t have children!');
  if not IsChild(P) then
  begin
    if (P^.Parent <> nil) then
      Error(P^.mName + ' already has another parents!');
    if Sex <> P^.Sex then
      Spouse^.HasChild(P)
    else
    begin
      if Child <> nil then
        P^.Sibling := Child;
      Child := P;
      P^.Parent := @Self;
    end;
  end;
  if (P^.mFSureName = '') and (P^.Father^.mFSureName <> '') then
    P^.mFSureName := P^.Father^.mFSureName;
end;


function  TPerson.Name : TName;
begin
  Name := mName;
end;

function  TPerson.FarsiName : TName;
begin
  FarsiName := mFName;
end;

function  TPerson.Id : TId;
begin
  Id := mId;
end;

function  TPerson.Sex : TSex;
begin
  Sex := mSex;
end;

function  TPerson.Formal : Boolean;
begin
  Formal := mFormal;
end;

function  TPerson.WriteName : string;
var
  S : string;
begin
  if mFName <> '' then
    S := mFName
  else
    S := mName;
  if mFSureName <> '' then
    if Farsi then
      S := mFSureName + ' ' + S
    else
      S := S + ' ' + mFSureName;
  WriteName := S;
end;

function  TPerson.IsMale : Boolean;
begin
  IsMale := mSex = Male;
end;

function  TPerson.IsFemale : Boolean;
begin
  IsFemale := not IsMale;
end;

function  TPerson.Son : PPerson;
begin
  if IsMale then Son := Child else
  if Spouse = nil then Son := nil else Son := Spouse^.Son;
end;

function  TPerson.Daughter : PPerson;
begin
  if IsFemale then Daughter := Child else
  if Spouse = nil then Daughter := nil else Daughter := Spouse^.Daughter;
end;

function  TPerson.Father : PPerson;
begin
  if IsMale then Father := Parent else
  if (Parent = nil) or (Parent^.Spouse = nil) then Father := nil else
  Father := Parent^.Spouse;
end;

function  TPerson.Mother : PPerson;
begin
  if IsFemale then Mother := Parent else
  if (Parent = nil) or (Parent^.Spouse = nil) then Mother := nil else
  Mother := Parent^.Spouse;
end;

function  TPerson.Husband : PPerson;
begin
  if IsFemale then Husband := Spouse else Husband := nil;
end;

function  TPerson.Wife : PPerson;
begin
  if IsMale then Wife := Spouse else Wife := nil;
end;


function  TPerson.IsParent (P : PPerson) : Boolean;
begin
  if Sex = P^.Sex then
    IsParent := Parent = P
  else
    IsParent := (Parent <> nil) and (Parent^.Spouse = P);
end;

function  TPerson.IsSpouse (P : PPerson) : Boolean;
begin
  IsSpouse := Spouse = P;
end;

function  TPerson.IsChild (P : PPerson) : Boolean;
begin
  IsChild := P^.IsParent(@Self);
end;

function  TPerson.IsSon (P : PPerson) : Boolean;
begin
  IsSon := P^.IsMale and IsChild(P);
end;

function  TPerson.IsDaughter (P : PPerson) : Boolean;
begin
  IsDaughter := P^.IsFemale and IsChild(P);
end;

function  TPerson.IsFather (P : PPerson) : Boolean;
begin
  IsFather := P^.IsMale and IsParent(P);
end;

function  TPerson.IsMother (P : PPerson) : Boolean;
begin
  IsMother := P^.IsFemale and IsParent(P);
end;

function  TPerson.IsHusband (P : PPerson) : Boolean;
begin
  IsHusband := P^.IsMale and IsSpouse(P);
end;

function  TPerson.IsWife (P : PPerson) : Boolean;
begin
  IsWife := P^.IsFemale and IsSpouse(P);
end;

procedure IncludePerson (var L : TList; P : PPerson);
begin
  if P <> nil then
    Include(L, P^.Id);
end;

procedure IncludeList (var L : TList; P : PPerson);
begin
  while P <> nil do
  begin
    Include(L, P^.Id);
    P := P^.Sibling;
  end;
end;

procedure TPerson.List (var L : TList; Action : Char);
begin
  Action := UpCase(Action);
  L := [];
  case Action of
    {Father}
    'F': IncludePerson(L, Father);
    {Mother}
    'M': IncludePerson(L, Mother);
    {Parents}
    'P': begin
    IncludePerson(L, Father);
    IncludePerson(L, Mother); end;

    {Wife}
    'W': IncludePerson(L, Wife);
    {Husband}
    'H': IncludePerson(L, Husband);
    {Spouse}
    'X': IncludePerson(L, Spouse);

    {Sons}
    'S': IncludeList(L, Son);
    {Daughters}
    'D': IncludeList(L, Daughter);
    {Children}
    'C': begin IncludeList(L, Son); IncludeList(L, Daughter); end;

    {Brothers}
    'B': if Father <> nil then
         begin IncludeList(L, Father^.Son); Exclude(L, mId); end;
    {Sisters}
    'T': if Father <> nil then
         begin IncludeList(L, Father^.Daughter); Exclude(L, mId); end;
    {Siblings}
    'L': if Father <> nil then
         begin IncludeList(L, Father^.Son); IncludeList(L, Father^.Daughter);
               Exclude(L, mId); end;
  end;
end;

end.
