{$A+,B-,D+,E-,F-,G+,I+,L+,N+,O-,P-,Q-,R-,S-,T-,V+,X+,Y+}
{by Behdad Esfahbod}
unit
  _Relation;

interface

  uses
    _Global, _Person;

  type
    PRelation = ^ TRelation;
    PRelationsArray = ^ TRelationsArray;
    TRelationsArray = array [1 .. MaxRelations] of PRelation;

    TRelation = object
    private
      mName : TName;
      mFarsiName : TName;
      mId : TId;
      mDef : TDef;
      mList : TList;
      mFormal : Boolean;

    public
      constructor Init (TheName, TheFarsiName : TName; TheId : TId; TheDef : TDef);

      procedure Reset;
      procedure Add (P : PPerson);
      function  InList (const P : PPerson) : Boolean;
      procedure SolveFor (const P : PPerson; const People : TPeopleArray; N : Integer);

      function  Name : TName;
      function  FarsiName : TName;
      function  Id : TId;
      function  Definition : TDef;
      function  Formal : Boolean;
    end;

implementation

constructor TRelation.Init (TheName, TheFarsiName : TName; TheId : TId; TheDef : TDef);
begin
  mFormal := TheName[1] <> '-';
  if TheName[1] = '-' then
    Delete(TheName, 1, 1);
  mName := TheName;
  if TheFarsiName <> '' then
    mFarsiName := TheFarsiName
  else
    mFarsiName := mName;
  mId := TheId;
  mDef := TheDef;
  Reset;
end;


procedure TRelation.Reset;
begin
  mList := [];
end;

procedure TRelation.Add (P : PPerson);
begin
  Include(mList, P^.Id);
end;

function  TRelation.InList (const P : PPerson) : Boolean;
begin
  InList := P^.Id in mList;
end;

procedure TRelation.SolveFor (const P : PPerson; const People : TPeopleArray; N : Integer);
var
  D : array [0 .. 1] of TList;
  T : TList;
  DN : Integer;
  I, J, K : Integer;
  Action : Char;
begin
  DN := 0;
  D[DN] := [P^.Id];
  for I := Length(mDef) downto 1 do
  begin
    DN := DN xor 1;
    D[DN] := [];
    Action := mDef[I];
    K := 0;
    for J := 1 to N do
      if J in D[DN xor 1] then
      begin
        Inc(K);
        People[J]^.List(T, Action);
        D[DN] := D[DN] + T;
        Exclude(D[DN xor 1], J);
        if D[DN xor 1] = [] then
          Break;
      end;
    if K = 0 then
      Break;
  end;
  Reset;
  if I = 1 then
    mList := D[DN];
end;


function  TRelation.Name : TName;
begin
  Name := mName;
end;

function  TRelation.FarsiName : TName;
begin
  FarsiName := mFarsiName;
end;

function  TRelation.Id : TId;
begin
  Id := mId;
end;

function  TRelation.Definition : TDef;
begin
  Definition := mDef;
end;

function  TRelation.Formal : Boolean;
begin
  Formal := mFormal;
end;

end.
